package com.css.simulation.resource.scheduler.scheduler;

import api.common.pojo.constants.DictConstants;
import api.common.pojo.dto.ProjectMessageDTO;
import api.common.util.CollectionUtil;
import api.common.util.JsonUtil;
import api.common.util.StringUtil;
import api.common.util.TimeUtil;
import com.css.simulation.resource.scheduler.consumer.ProjectConsumer;
import com.css.simulation.resource.scheduler.mapper.*;
import com.css.simulation.resource.scheduler.pojo.po.ClusterPO;
import com.css.simulation.resource.scheduler.pojo.po.ProjectPO;
import com.css.simulation.resource.scheduler.pojo.po.TaskPO;
import com.css.simulation.resource.scheduler.pojo.po.UserPO;
import com.css.simulation.resource.scheduler.pojo.to.PrefixTO;
import com.css.simulation.resource.scheduler.service.TaskService;
import com.css.simulation.resource.scheduler.util.ProjectUtil;
import io.kubernetes.client.openapi.ApiClient;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

@Component
@Slf4j
public class ProjectScheduler {

    @Value("${scheduler.host.hostname}")
    String hostname;
    @Value("${scheduler.host.username}")
    String username;
    @Value("${scheduler.host.password}")
    String password;
    @Value("${scheduler.linux-path.job-yaml}")
    String jobYaml;
    // -------------------------------- Comment --------------------------------
    @Resource
    StringRedisTemplate stringRedisTemplate;
    @Resource
    TaskService taskService;
    @Resource
    TaskMapper taskMapper;
    @Resource
    ClusterMapper clusterMapper;
    @Resource
    ManualProjectMapper manualProjectMapper;
    @Resource
    AutoSubProjectMapper autoSubProjectMapper;
    @Resource
    ApiClient apiClient;
    @Resource
    ProjectConsumer projectConsumer;
    @Resource
    ProjectUtil projectUtil;
    @Resource
    UserMapper userMapper;


    /**
     * 调度项目启动
     */
    @Scheduled(fixedDelay = 60 * 1000)
    @SneakyThrows
    public void dispatchProject() {

        //1 查询已经在页面上点击运行的项目(后面会判断是排队中还是已经与运行了)
        List<ProjectPO> manualProjectList = manualProjectMapper.selectByNowRunState(DictConstants.PROJECT_RUNNING);
        List<ProjectPO> autoSubProjectList = autoSubProjectMapper.selectByNowRunState(DictConstants.PROJECT_RUNNING);
        List<ProjectPO> allProject = new ArrayList<>();
        allProject.addAll(manualProjectList);
        allProject.addAll(autoSubProjectList);


        for (ProjectPO project : allProject) {
            String projectId = project.getId();
            long parallelism = Long.parseLong(project.getParallelism());
            String userId = project.getCreateUserId();
            UserPO userPO = userMapper.selectById(userId);
            String roleCode = userPO.getRoleCode();
            String useType = userPO.getUseType();
            ClusterPO clusterPO = null;
            String clusterId;
            boolean isSystem = false;
            if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
                clusterId = DictConstants.SYSTEM_CLUSTER_ID;
                isSystem = true;
            } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
                clusterPO = clusterMapper.selectByUserId(userId);
                if (clusterPO == null) {
                    log.error("ProjectScheduler--dispatchProject 项目 " + projectId + " 的创建用户 " + userId + " 没有分配集群！");
                    return;
                }
                clusterId = clusterPO.getId();
            } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
                if (DictConstants.USE_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                    clusterPO = clusterMapper.selectByUserId(userId);
                } else {    //3-4 共享子账户，根据父账户的共享节点排队
                    String parentUserId = userPO.getCreateUserId();
                    clusterPO = clusterMapper.selectByUserId(parentUserId);
                }
                if (clusterPO == null) {
                    log.error("ProjectScheduler--dispatchProject 项目 " + projectId + " 的创建用户 " + userId + " 没有分配集群！");
                    return;
                }
                clusterId = clusterPO.getId();
            } else {
                log.error("ProjectConsumer--dispatchProject 项目 " + projectId + " 的创建人 " + userId + " 为未知账户类型，不予执行！");
                return;
            }
            PrefixTO redisPrefix = projectUtil.getRedisPrefixByClusterIdAndProjectId(clusterId, projectId);
            // --------------------------------  判断项目是否已经在执行，如果执行则 continue --------------------------------
            if (StringUtil.isNotEmpty(stringRedisTemplate.opsForValue().get(redisPrefix.getProjectRunningKey()))) {
                continue;
            }
            // -------------------------------- 项目没有执行说明等待中 --------------------------------
            if (isSystem) { // 系统管理员直接执行
                run(clusterId, projectId, redisPrefix.getProjectWaitingKey(), redisPrefix.getProjectRunningKey(), parallelism);
                return;
            }
            int simulationLicenseNumber = clusterPO.getNumSimulationLicense();
            // 获取该用户正在运行的项目数量
            Set<String> clusterRunningKeySet = stringRedisTemplate.keys(redisPrefix.getClusterRunningPrefix() + "*");
            List<String> runningProjectSet = null;
            // cluster:${clusterId}:running:${projectId}
            if (CollectionUtil.isNotEmpty(clusterRunningKeySet)) {
                runningProjectSet = projectUtil.getRunningProjectList(clusterRunningKeySet);
                if (CollectionUtil.isNotEmpty(runningProjectSet)) {
                    log.info("ProjectScheduler--dispatchProject 运行中的项目的 key 有：" + runningProjectSet);
                    long parallelismSum = 0;
                    for (String runningProjectKey : runningProjectSet) {
                        parallelismSum += JsonUtil.jsonToBean(stringRedisTemplate.opsForValue().get(runningProjectKey), ProjectMessageDTO.class).getParallelism();
                    }
                    if (parallelismSum < simulationLicenseNumber) {
                        if (parallelismSum + parallelism < simulationLicenseNumber) {
                            run(clusterId, projectId, redisPrefix.getProjectWaitingKey(), redisPrefix.getProjectRunningKey(), parallelism);
                            return;
                        }
                    }
                }
            }
            if ((CollectionUtil.isEmpty(clusterRunningKeySet) || CollectionUtil.isEmpty(runningProjectSet)) && parallelism < simulationLicenseNumber) {
                run(clusterId, projectId, redisPrefix.getProjectWaitingKey(), redisPrefix.getProjectRunningKey(), parallelism);
            }
        }
    }


    @SneakyThrows
    public void run(String clusterId, String projectId, String projectWaitingKey, String projectRunningKey, long parallelism) {
        ProjectMessageDTO projectMessageDTO = JsonUtil.jsonToBean(stringRedisTemplate.opsForValue().get(projectWaitingKey), ProjectMessageDTO.class);

        //1 获取所有节点的剩余可用并行度
        Map<String, Integer> nodeMap = projectUtil.getNodeMapToUse(parallelism);
        if (CollectionUtil.isEmpty(nodeMap)) {
            return;
        }
        //2 计算实际可用并行度
        int parallelismSum = nodeMap.keySet().stream().mapToInt(nodeMap::get).sum();

        //2 判断剩余可用并行度是否大于项目并行度，否则加入扩充队列
        if (parallelismSum > 0L) {
            log.info("ProjectConsumer--run 集群 " + clusterId + " 将项目 " + projectId + "在节点 " + nodeMap + " 上以并行度 " + parallelismSum + " 执行！");
            projectMessageDTO.setCurrentParallelism(parallelismSum);    // 设置实际的并行度
            projectConsumer.parseProject(nodeMap, projectMessageDTO, "cluster:" + clusterId, projectRunningKey);
        }
    }


    /**
     * 处理 pod 超时
     * 同时也可处理 pod 莫名关闭，因为关闭之后也会超时
     */
    @Scheduled(fixedDelay = 60 * 1000)
    public void taskTimeout() {
        long timeout = 2 * 60 * 1000L;
        List<TaskPO> executingTaskList = taskMapper.selectByRunState(DictConstants.TASK_RUNNING);
        if (executingTaskList != null && executingTaskList.size() > 0) {
            log.info("ProjectScheduler--taskTimeout 正在运行的任务有：" + executingTaskList);
            for (TaskPO task : executingTaskList) {
                String userId = task.getCreateUserId();
                String projectId = task.getPId();
                String taskId = task.getId();
                PrefixTO redisPrefix = projectUtil.getRedisPrefixByUserIdAndProjectIdAndTaksId(userId, projectId, taskId);
                // 获取心跳时间
                String tickTime = stringRedisTemplate.opsForValue().get(redisPrefix.getTaskTickKey());
                if (StringUtil.isEmpty(tickTime)) {
                    log.error(redisPrefix.getTaskTickKey() + "，该 key 的心跳时间为空！");
                    continue;
                }
                long lastTickTime = Long.parseLong(tickTime);
                // 如果心跳超时则更改任务状态为 Aborted
                if (TimeUtil.getNow() - lastTickTime > timeout) {
                    String podName = stringRedisTemplate.opsForValue().get(redisPrefix.getTaskPodKey());
                    taskService.taskState(taskId, DictConstants.TASK_ABORTED, podName);
                }
            }
        }
    }

}
